@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2016-2020, OARC, Inc.
+ * Copyright (c) 2016-2025 OARC, Inc.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -116,10 +116,10 @@ typedef int filter_t(const char* descr,
     unsigned                     sport,
     unsigned                     dport,
     my_bpftimeval                ts,
-    const u_char*                pkt_copy,
-    const unsigned               olen,
-    const u_char*                payload,
-    const unsigned               payloadlen);
+    u_char*                      pkt_copy,
+    unsigned                     olen,
+    u_char*                      payload,
+    unsigned                     payloadlen);
 
 /*
  * Extensions
@@ -140,6 +140,9 @@ typedef void (*tcpstate_reset_t)(void* tcpstate, const char* msg);
 #define DNSCAP_EXT_SET_IADDR 5
 typedef void (*set_iaddr_t)(iaddr* from, iaddr* to);
 
+#define DNSCAP_EXT_SET_OUTPUT_PKT 6
+typedef void (*set_output_pkt_t)(u_char* pkt, const unsigned olen);
+
 /*
  * Flags
  */
