@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2016-2020, OARC, Inc.
+ * Copyright (c) 2016-2025 OARC, Inc.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -42,6 +42,9 @@
 #include "endpoint.h"
 #include "tcpstate.h"
 #include "tcpreasm.h"
+#include "endian_compat.h"
+
+#include <ldns/ldns.h>
 
 struct ip6_hdr* network_ipv6 = 0;
 struct ip*      network_ip   = 0;
@@ -49,15 +52,19 @@ struct udphdr*  network_udp  = 0;
 
 extern tcpstate_ptr _curr_tcpstate; /* from tcpstate.c */
 
-#define MY_GET32(l, cp)                                          \
-    do {                                                         \
-        register const u_char* t_cp = (const u_char*)(cp);       \
-        (l)                         = ((u_int32_t)t_cp[0] << 24) \
-              | ((u_int32_t)t_cp[1] << 16)                       \
-              | ((u_int32_t)t_cp[2] << 8)                        \
-              | ((u_int32_t)t_cp[3]);                            \
-        (cp) += NS_INT32SZ;                                      \
-    } while (0)
+static inline uint16_t _need16(const void* ptr)
+{
+    uint16_t v;
+    memcpy(&v, ptr, sizeof(v));
+    return be16toh(v);
+}
+
+static inline uint32_t _need32(const void* ptr)
+{
+    uint32_t v;
+    memcpy(&v, ptr, sizeof(v));
+    return be32toh(v);
+}
 
 static int skip_vlan(unsigned vlan)
 {
@@ -189,34 +196,50 @@ void layer_pkt(u_char* user, const pcap_thread_packet_t* packet, const u_char* p
         gmtime_r(&t, &tm);
         strftime(when, sizeof(when), "%Y-%m-%d %T", &tm);
 
+        if (packet->have_iphdr && !options.defrag_ipv4) {
+            if (len > packet->iphdr.ip_len) {
+                len = packet->iphdr.ip_len;
+            }
+        } else if (packet->have_ip6hdr && !options.defrag_ipv6) {
+            if (len > 40 + packet->ip6hdr.ip6_plen) {
+                len = 40 + packet->ip6hdr.ip6_plen;
+            }
+        }
+
         if (vlan != MAX_VLAN) {
-            snprintf(descr, sizeof(descr), "[%lu] %s.%06lu [#%ld %s (vlan %u) %u] \\\n",
-                (u_long)len,
+            snprintf(descr, sizeof(descr), "[%zu] %s.%06ld [#%zd %s (vlan %u) %u] \\\n",
+                len,
                 when,
-                (u_long)firstpkt->pkthdr.ts.tv_usec,
-                (long)msgcount,
+                firstpkt->pkthdr.ts.tv_usec,
+                msgcount,
                 mypcap->name ? mypcap->name : "\"some interface\"",
                 vlan,
                 vlan);
         } else {
-            snprintf(descr, sizeof(descr), "[%lu] %s.%06lu [#%ld %s %u] \\\n",
-                (u_long)len,
+            snprintf(descr, sizeof(descr), "[%zu] %s.%06ld [#%zd %s %u] \\\n",
+                len,
                 when,
-                (u_long)firstpkt->pkthdr.ts.tv_usec,
-                (long)msgcount,
+                firstpkt->pkthdr.ts.tv_usec,
+                msgcount,
                 mypcap->name ? mypcap->name : "\"some interface\"",
                 vlan);
         }
     }
 
-    if (next_interval != 0 && firstpkt->pkthdr.ts.tv_sec >= next_interval && dumper_opened == dump_state)
-        dumper_close(firstpkt->pkthdr.ts);
+    if (next_interval != 0 && firstpkt->pkthdr.ts.tv_sec >= next_interval) {
+        if (preso)
+            goto breakloop;
+        if (dumper_opened == dump_state)
+            dumper_close(firstpkt->pkthdr.ts);
+        if (dump_type == to_stdout)
+            goto breakloop;
+    }
     if (dumper_closed == dump_state && dumper_open(firstpkt->pkthdr.ts))
         goto breakloop;
 
     network_pkt2(descr, firstpkt->pkthdr.ts, packet, payload, length);
 
-    if (limit_packets != 0U && msgcount == limit_packets) {
+    if (limit_packets != 0U && msgcount >= limit_packets) {
         if (preso)
             goto breakloop;
         if (dumper_opened == dump_state && dumper_close(firstpkt->pkthdr.ts))
@@ -269,32 +292,33 @@ void dl_pkt(u_char* user, const struct pcap_pkthdr* hdr, const u_char* pkt, cons
     case DLT_NULL: {
         uint32_t x;
 
-        if (len < NS_INT32SZ)
+        if (len < 4)
             return;
-        x = *(const uint32_t*)pkt;
+        x = _need32(pkt);
         if (x == PF_INET)
             etype = ETHERTYPE_IP;
         else if (x == PF_INET6)
             etype = ETHERTYPE_IPV6;
         else
             return;
-        pkt += NS_INT32SZ;
-        len -= NS_INT32SZ;
+        pkt += 4;
+        len -= 4;
         break;
     }
     case DLT_LOOP: {
         uint32_t x;
 
-        if (len < NS_INT32SZ)
+        if (len < 4)
             return;
-        MY_GET32(x, pkt);
-        len -= NS_INT32SZ;
+        x = _need32(pkt);
         if (x == PF_INET)
             etype = ETHERTYPE_IP;
         else if (x == PF_INET6)
             etype = ETHERTYPE_IPV6;
         else
             return;
+        pkt += 4;
+        len -= 4;
         break;
     }
     case DLT_RAW: {
@@ -324,10 +348,10 @@ void dl_pkt(u_char* user, const struct pcap_pkthdr* hdr, const u_char* pkt, cons
         if (etype == ETHERTYPE_VLAN) {
             if (len < 4)
                 return;
-            vlan = ntohs(*(const uint16_t*)pkt) & 0xFFF;
+            vlan = _need16(pkt) & 0xFFF;
             pkt += 2;
             len -= 2;
-            etype = ntohs(*(const uint16_t*)pkt);
+            etype = _need16(pkt);
             pkt += 2;
             len -= 2;
         }
@@ -337,11 +361,21 @@ void dl_pkt(u_char* user, const struct pcap_pkthdr* hdr, const u_char* pkt, cons
     case DLT_LINUX_SLL: {
         if (len < 16)
             return;
-        etype = ntohs(*(const uint16_t*)&pkt[14]);
+        etype = _need16(&pkt[14]);
         pkt += 16;
         len -= 16;
         break;
     }
+#endif
+#ifdef DLT_LINUX_SLL2
+    case DLT_LINUX_SLL2: {
+        if (len < 20)
+            return;
+        etype = _need16(&pkt[0]);
+        pkt += 20;
+        len -= 20;
+        break;
+    }
 #endif
     default:
         return;
@@ -403,20 +437,26 @@ void dl_pkt(u_char* user, const struct pcap_pkthdr* hdr, const u_char* pkt, cons
         } else {
             viap = "\"some interface\"";
         }
-        snprintf(descr, sizeof(descr), "[%lu] %s.%06lu [#%ld %s %u] \\\n",
-            (u_long)len, when, (u_long)hdr->ts.tv_usec, (long)msgcount, viap, vlan);
+        snprintf(descr, sizeof(descr), "[%zu] %s.%06ld [#%zu %s %u] \\\n",
+            len, when, hdr->ts.tv_usec, msgcount, viap, vlan);
     } else {
         descr[0] = '\0';
     }
 
-    if (next_interval != 0 && hdr->ts.tv_sec >= next_interval && dumper_opened == dump_state)
-        dumper_close(hdr->ts);
+    if (next_interval != 0 && hdr->ts.tv_sec >= next_interval) {
+        if (preso)
+            goto breakloop;
+        if (dumper_opened == dump_state)
+            dumper_close(hdr->ts);
+        if (dump_type == to_stdout)
+            goto breakloop;
+    }
     if (dumper_closed == dump_state && dumper_open(hdr->ts))
         goto breakloop;
 
     network_pkt(descr, hdr->ts, pf, pkt, len);
 
-    if (limit_packets != 0U && msgcount == limit_packets) {
+    if (limit_packets != 0U && msgcount >= limit_packets) {
         if (preso)
             goto breakloop;
         if (dumper_opened == dump_state && dumper_close(hdr->ts))
@@ -440,17 +480,164 @@ breakloop:
     main_exit = TRUE;
 }
 
+/*
+ * Determine if packet should be filtered based on query type
+ * return values: -1 error
+ *                 0 don't filter
+ *                 1 filter
+ * for non-zero return status, reason should be set to a
+ * static string.
+ */
+static int
+_filter_by_qtype(const ldns_pkt* lpkt, char** reason)
+{
+    ldns_rr_list* rrs = ldns_pkt_question(lpkt);
+    if (!rrs) {
+        *reason = "failed to get list of questions";
+        return -1;
+    }
+    /* Look at each RR in the section (or each QNAME in
+       the question section). */
+    size_t i, n;
+    for (i = 0, n = ldns_rr_list_rr_count(rrs); i < n; i++) {
+        ldns_rr* rr = ldns_rr_list_rr(rrs, i);
+        if (!rr) {
+            *reason = "failed to get question";
+            return -1;
+        }
+        if (match_qtype && ldns_rr_get_type(rr) == match_qtype) {
+            if (dumptrace >= 2)
+                fprintf(stderr, "Packet query type %u matches filter on type == %u\n", ldns_rr_get_type(rr), match_qtype);
+            return 0;
+        } else if (nmatch_qtype && ldns_rr_get_type(rr) == nmatch_qtype) {
+            if (dumptrace >= 2)
+                fprintf(stderr, "Packet query type %u matches filter on type != %u\n", ldns_rr_get_type(rr), nmatch_qtype);
+            *reason = "matched unwanted qtype";
+            return 1;
+        }
+    }
+    if (match_qtype) {
+        *reason = "didn't match wanted qtype";
+        return 1; // didn't match any question RRs
+    }
+    return 0;
+}
+
+/*
+ * Determine if packet should be filtered based on query name
+ * return values: -1 error
+ *                 0 don't filter
+ *                 1 filter
+ * for non-zero return status, reason should be set to a
+ * static string.
+ */
+static int
+_match_rr(const ldns_rr_list* rrs, char** reason, int* negmatch, int* match, ldns_buffer* buf)
+{
+    /* Look at each RR in the section (or each QNAME in
+       the question section). */
+    myregex_ptr myregex;
+    size_t      i, n;
+    for (i = 0, n = ldns_rr_list_rr_count(rrs); i < n; i++) {
+        ldns_rr* rr = ldns_rr_list_rr(rrs, i);
+        if (!rr) {
+            *reason = "failed to get RR";
+            return -1;
+        }
+
+        ldns_buffer_clear(buf);
+        if (ldns_rdf2buffer_str(buf, ldns_rr_owner(rr)) != LDNS_STATUS_OK) {
+            *reason = "failed to get RR";
+            return -1;
+        }
+
+        for (myregex = HEAD(myregexes);
+             myregex != NULL;
+             myregex = NEXT(myregex, link)) {
+
+            if (regexec(&myregex->reg, (char*)ldns_buffer_begin(buf), 0, NULL, 0) == 0) {
+                if (myregex->not )
+                    (*negmatch)++;
+                else
+                    (*match)++;
+
+                if (dumptrace >= 2)
+                    fprintf(stderr,
+                        "; \"%s\" %s~ /%s/ %d %d\n",
+                        (char*)ldns_buffer_begin(buf),
+                        myregex->not ? "!" : "",
+                        myregex->str,
+                        *match,
+                        *negmatch);
+            }
+        }
+    }
+    return 0;
+}
+
+static int
+_filter_by_qname(const ldns_pkt* lpkt, char** reason)
+{
+    int          match    = -1;
+    int          negmatch = -1;
+    ldns_buffer* buf      = ldns_buffer_new(512);
+    if (!buf) {
+        fprintf(stderr, "%s: out of memory", ProgramName);
+        exit(1);
+    }
+
+    /*
+     * Initialize matching counters
+     */
+    myregex_ptr myregex;
+    for (myregex = HEAD(myregexes); myregex != NULL; myregex = NEXT(myregex, link)) {
+        if (myregex->not ) {
+            negmatch = 0;
+        } else {
+            match = 0;
+        }
+    }
+
+    ldns_rr_list* rrs;
+    if ((rrs = ldns_pkt_question(lpkt))) {
+        if (_match_rr(rrs, reason, &negmatch, &match, buf))
+            return -1;
+    }
+    if ((rrs = ldns_pkt_answer(lpkt))) {
+        if (_match_rr(rrs, reason, &negmatch, &match, buf))
+            return -1;
+    }
+    if ((rrs = ldns_pkt_authority(lpkt))) {
+        if (_match_rr(rrs, reason, &negmatch, &match, buf))
+            return -1;
+    }
+    if ((rrs = ldns_pkt_additional(lpkt))) {
+        if (_match_rr(rrs, reason, &negmatch, &match, buf))
+            return -1;
+    }
+    ldns_buffer_free(buf);
+
+    /*
+     * Fail if any negative matches or if no positive matches
+     */
+    if (negmatch > 0 || match == 0) {
+        *reason = "failed regex match";
+        return 1;
+    }
+    return 0;
+}
+
 void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_t* packet, const u_char* payload, size_t length)
 {
-    u_char        pkt_copy[SNAPLEN], *pkt = pkt_copy;
-    const u_char* dnspkt = 0;
-    unsigned      proto, sport, dport;
-    iaddr         from, to, initiator, responder;
-    int           response, m;
-    unsigned      flags    = DNSCAP_OUTPUT_ISLAYER;
-    tcpstate_ptr  tcpstate = NULL;
-    size_t        len, dnslen = 0;
-    HEADER        dns;
+    u_char       pkt_copy[SNAPLEN], *pkt = pkt_copy;
+    u_char*      dnspkt = 0;
+    unsigned     proto, sport, dport;
+    iaddr        from, to, initiator, responder;
+    int          response, m;
+    unsigned     flags    = DNSCAP_OUTPUT_ISLAYER;
+    tcpstate_ptr tcpstate = NULL;
+    size_t       len, dnslen = 0;
+    HEADER       dns;
 
     /* Make a writable copy of the packet and use that copy from now on. */
     if (length > SNAPLEN)
@@ -497,8 +684,8 @@ void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_
         proto  = IPPROTO_UDP;
         sport  = packet->udphdr.uh_sport;
         dport  = packet->udphdr.uh_dport;
-        dnspkt = payload;
-        dnslen = length;
+        dnspkt = pkt;
+        dnslen = len;
         flags |= DNSCAP_OUTPUT_ISDNS;
     } else if (packet->have_tcphdr) {
         uint32_t seq = packet->tcphdr.th_seq;
@@ -538,11 +725,11 @@ void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_
 
         tcpstate = tcpstate_find(from, to, sport, dport, ts.tv_sec);
         if (dumptrace >= 3) {
-            fprintf(stderr, "%s: tcp pkt: %lu.%06lu [%4lu] %15s -> ",
+            fprintf(stderr, "%s: tcp pkt: %" PRIdMAX ".%06ld [%4zu] %15s -> ",
                 ProgramName,
-                (u_long)ts.tv_sec,
-                (u_long)ts.tv_usec,
-                (u_long)len,
+                (intmax_t)ts.tv_sec,
+                ts.tv_usec,
+                len,
                 ia_str(from));
             fprintf(stderr, "%15s; ", ia_str(to));
 
@@ -563,14 +750,7 @@ void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_
             _curr_tcpstate = 0;
 
             /* End of stream; deallocate the tcpstate. */
-            if (tcpstate) {
-                UNLINK(tcpstates, tcpstate, link);
-                if (tcpstate->reasm) {
-                    tcpreasm_free(tcpstate->reasm);
-                }
-                free(tcpstate);
-                tcpstate_count--;
-            }
+            tcpstate_free(tcpstate);
             return;
         }
         if (packet->tcphdr.th_flags & TH_SYN) {
@@ -593,7 +773,7 @@ void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_
                     fprintf(stderr, "warning: recycling state for "
                         "duplicate tcp stream after only %ld "
                         "seconds idle\n",
-                        (u_long)(ts.tv_sec - tcpstate->last_use));
+                        ts.tv_sec - tcpstate->last_use);
                     */
                 }
             } else {
@@ -811,6 +991,10 @@ void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_
         }
 
         /* Application. */
+        if (!dnspkt) {
+            tcpstate_discard(tcpstate, "no dns");
+            return;
+        }
         if (dnslen < sizeof dns) {
             tcpstate_discard(tcpstate, "too small");
             return;
@@ -846,7 +1030,7 @@ void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_
             tcpstate_discard(tcpstate, "missing required host");
             return;
         }
-        if (!(((msg_wanted & MSG_QUERY) != 0 && dns.opcode == ns_o_query) || ((msg_wanted & MSG_UPDATE) != 0 && dns.opcode == ns_o_update) || ((msg_wanted & MSG_NOTIFY) != 0 && dns.opcode == ns_o_notify))) {
+        if (!(((msg_wanted & MSG_QUERY) != 0 && dns.opcode == LDNS_PACKET_QUERY) || ((msg_wanted & MSG_UPDATE) != 0 && dns.opcode == LDNS_PACKET_UPDATE) || ((msg_wanted & MSG_NOTIFY) != 0 && dns.opcode == LDNS_PACKET_NOTIFY))) {
             tcpstate_discard(tcpstate, "unwanted opcode");
             return;
         }
@@ -863,19 +1047,13 @@ void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_
                 return;
             }
         }
-#if HAVE_NS_INITPARSE && HAVE_NS_PARSERR && HAVE_NS_SPRINTRR
-        if (!EMPTY(myregexes)) {
-            int     match, negmatch;
-            ns_msg  msg;
-            ns_sect s;
-
-            match    = -1;
-            negmatch = -1;
-            if (ns_initparse(dnspkt, dnslen, &msg) < 0) {
+        if (!EMPTY(myregexes) || match_qtype || nmatch_qtype) {
+            ldns_pkt* lpkt = 0;
+            if (ldns_wire2pkt(&lpkt, dnspkt, dnslen) != LDNS_STATUS_OK) {
                 /* DNS message may have padding, try get actual size */
-                if (errno == EMSGSIZE) {
-                    size_t dnslen2 = calcdnslen(dnspkt, dnslen);
-                    if (dnslen2 > 0 && dnslen2 < dnslen && ns_initparse(dnspkt, dnslen2, &msg) < 0) {
+                size_t dnslen2 = calcdnslen(dnspkt, dnslen);
+                if (dnslen2 > 0 && dnslen2 < dnslen) {
+                    if (ldns_wire2pkt(&lpkt, dnspkt, dnslen2) != LDNS_STATUS_OK) {
                         tcpstate_discard(tcpstate, "failed parse");
                         return;
                     }
@@ -884,74 +1062,19 @@ void network_pkt2(const char* descr, my_bpftimeval ts, const pcap_thread_packet_
                     return;
                 }
             }
-            /* Look at each section of the message:
-                 question, answer, authority, additional */
-            for (s = ns_s_qd; s < ns_s_max; s++) {
-                char        pres[SNAPLEN * 4];
-                const char* look;
-                int         count, n;
-                ns_rr       rr;
-
-                /* Look at each RR in the section (or each QNAME in
-                   the question section). */
-                count = ns_msg_count(msg, s);
-                for (n = 0; n < count; n++) {
-                    myregex_ptr myregex;
-
-                    if (ns_parserr(&msg, s, n, &rr) < 0) {
-                        tcpstate_discard(tcpstate, "failed parse");
-                        return;
-                    }
-                    if (s == ns_s_qd) {
-                        look = ns_rr_name(rr);
-                    } else {
-                        if (ns_sprintrr(&msg, &rr, NULL, ".",
-                                pres, sizeof pres)
-                            < 0) {
-                            tcpstate_discard(tcpstate, "failed parse");
-                            return;
-                        }
-                        look = pres;
-                    }
-                    for (myregex = HEAD(myregexes);
-                         myregex != NULL;
-                         myregex = NEXT(myregex, link)) {
-                        if (myregex->not ) {
-                            if (negmatch < 0)
-                                negmatch = 0;
-                        } else {
-                            if (match < 0)
-                                match = 0;
-                        }
-
-                        if (regexec(&myregex->reg, look, 0, NULL, 0) == 0) {
-                            if (myregex->not )
-                                negmatch++;
-                            else
-                                match++;
-
-                            if (dumptrace >= 2)
-                                fprintf(stderr,
-                                    "; \"%s\" %s~ /%s/ %d %d\n",
-                                    look,
-                                    myregex->not ? "!" : "",
-                                    myregex->str,
-                                    match,
-                                    negmatch);
-                        }
-                    }
-                }
-            }
-            /*
-             * Fail if any negative matching or if no match, match can be -1 which
-             * indicates that there are only negative matching
-             */
-            if (negmatch > 0 || match == 0) {
-                tcpstate_discard(tcpstate, "failed regex match");
+            char* reason = 0;
+            if ((match_qtype || nmatch_qtype) && _filter_by_qtype(lpkt, &reason)) {
+                ldns_pkt_free(lpkt);
+                tcpstate_discard(tcpstate, reason);
                 return;
             }
+            if (!EMPTY(myregexes) && _filter_by_qname(lpkt, &reason)) {
+                ldns_pkt_free(lpkt);
+                tcpstate_discard(tcpstate, reason);
+                return;
+            }
+            ldns_pkt_free(lpkt);
         }
-#endif /* HAVE_NS_INITPARSE && HAVE_NS_PARSERR && HAVE_NS_SPRINTRR */
 
         /*
          * TODO: Policy hiding.
@@ -974,7 +1097,7 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
     const u_char* opkt, size_t olen)
 {
     u_char          pkt_copy[SNAPLEN], *pkt = pkt_copy;
-    const u_char*   dnspkt = 0;
+    u_char*         dnspkt = 0;
     unsigned        proto, sport, dport;
     iaddr           from, to, initiator, responder;
     struct ip6_hdr* ipv6;
@@ -986,6 +1109,7 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
     struct ip*      ip;
     size_t          len, dnslen = 0;
     HEADER          dns;
+    char            descr_[512];
 
     if (dumptrace >= 4)
         fprintf(stderr, "processing %s packet: len=%zu\n", (pf == PF_INET ? "IPv4" : (pf == PF_INET6 ? "IPv6" : "unknown")), olen);
@@ -1015,8 +1139,16 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
         to.af = AF_INET;
         memcpy(&to.u.a4, &ip->ip_dst, sizeof(struct in_addr));
         offset = ip->ip_hl << 2;
-        if (len > ntohs(ip->ip_len)) /* small IP packets have L2 padding */
+        if (len > ntohs(ip->ip_len)) {
+            /* IP packets have L2 padding */
+            olen -= len - ntohs(ip->ip_len); // remove ether frame padding
+            if (preso) {
+                // rewrite descr
+                snprintf(descr_, sizeof(descr_), "[%zu]%s", olen, strchr(descr, ' '));
+                descr = descr_;
+            }
             len = ntohs(ip->ip_len);
+        }
         if (len <= (size_t)offset)
             goto network_pkt_end;
         pkt += offset;
@@ -1048,6 +1180,16 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
         offset      = sizeof(struct ip6_hdr);
         payload_len = ntohs(ipv6->ip6_plen);
 
+        if (len > sizeof(*ipv6) + payload_len) {
+            /* IP packets have L2 padding */
+            olen -= len - sizeof(*ipv6) - payload_len; // remove ether frame padding
+            if (preso) {
+                // rewrite descr
+                snprintf(descr_, sizeof(descr_), "[%zu]%s", olen, strchr(descr, ' '));
+                descr = descr_;
+            }
+        }
+
         memset(&from, 0, sizeof from);
         from.af = AF_INET6;
         memcpy(&from.u.a6, &ipv6->ip6_src, sizeof(struct in6_addr));
@@ -1185,8 +1327,8 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
 
         tcpstate = tcpstate_find(from, to, sport, dport, ts.tv_sec);
         if (dumptrace >= 3) {
-            fprintf(stderr, "%s: tcp pkt: %lu.%06lu [%4lu] ", ProgramName,
-                (u_long)ts.tv_sec, (u_long)ts.tv_usec, (u_long)len);
+            fprintf(stderr, "%s: tcp pkt: %" PRIdMAX ".%06ld [%4zu] ", ProgramName,
+                (intmax_t)ts.tv_sec, ts.tv_usec, len);
             fprintf(stderr, "%15s -> ", ia_str(from));
             fprintf(stderr, "%15s; ", ia_str(to));
             if (tcpstate)
@@ -1204,14 +1346,7 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
                 pkt_copy, olen, NULL, 0);
             _curr_tcpstate = 0;
             /* End of stream; deallocate the tcpstate. */
-            if (tcpstate) {
-                UNLINK(tcpstates, tcpstate, link);
-                if (tcpstate->reasm) {
-                    tcpreasm_free(tcpstate->reasm);
-                }
-                free(tcpstate);
-                tcpstate_count--;
-            }
+            tcpstate_free(tcpstate);
             goto network_pkt_end;
         }
         if (tcp->th_flags & TH_SYN) {
@@ -1229,7 +1364,7 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
                 fprintf(stderr, "warning: recycling state for "
                     "duplicate tcp stream after only %ld "
                     "seconds idle\n",
-                    (u_long)(ts.tv_sec - tcpstate->last_use));
+                    ts.tv_sec - tcpstate->last_use);
             }
 #endif
             } else {
@@ -1338,8 +1473,8 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
                 tcpstate->lastdns = seq + tcpstate->dnslen;
             } else if (seqdiff == 0 && len > 2) {
                 /* This is the first segment of the stream, and
-             * contains the dnslen and dns header, so we can
-             * filter on it. */
+                 * contains the dnslen and dns header, so we can
+                 * filter on it. */
                 if (dumptrace >= 3)
                     fprintf(stderr, "len+hdr\n");
                 dnslen = tcpstate->dnslen = (pkt[0] << 8) | (pkt[1] << 0);
@@ -1351,7 +1486,7 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
                 tcpstate->lastdns = seq + 2 + tcpstate->dnslen;
             } else if (seqdiff == 0 && len == 2) {
                 /* This is the first segment of the stream, but only
-             * contains the dnslen. */
+                 * contains the dnslen. */
                 if (dumptrace >= 3)
                     fprintf(stderr, "len\n");
                 tcpstate->dnslen  = (pkt[0] << 8) | (pkt[1] << 0);
@@ -1368,7 +1503,7 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
                 goto network_pkt_end;
             } else if (seqdiff == 2) {
                 /* This is not the first segment, but it does contain
-             * the first dns header, so we can filter on it. */
+                 * the first dns header, so we can filter on it. */
                 if (dumptrace >= 3)
                     fprintf(stderr, "hdr\n");
                 tcpstate->maxdiff = seqdiff + (uint32_t)len;
@@ -1434,6 +1569,10 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
         }
 
         /* Application. */
+        if (!dnspkt) {
+            tcpstate_discard(tcpstate, "no dns");
+            goto network_pkt_end;
+        }
         if (dnslen < sizeof dns) {
             tcpstate_discard(tcpstate, "too small");
             goto network_pkt_end;
@@ -1469,7 +1608,7 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
             tcpstate_discard(tcpstate, "missing required host");
             goto network_pkt_end;
         }
-        if (!(((msg_wanted & MSG_QUERY) != 0 && dns.opcode == ns_o_query) || ((msg_wanted & MSG_UPDATE) != 0 && dns.opcode == ns_o_update) || ((msg_wanted & MSG_NOTIFY) != 0 && dns.opcode == ns_o_notify))) {
+        if (!(((msg_wanted & MSG_QUERY) != 0 && dns.opcode == LDNS_PACKET_QUERY) || ((msg_wanted & MSG_UPDATE) != 0 && dns.opcode == LDNS_PACKET_UPDATE) || ((msg_wanted & MSG_NOTIFY) != 0 && dns.opcode == LDNS_PACKET_NOTIFY))) {
             tcpstate_discard(tcpstate, "unwanted opcode");
             goto network_pkt_end;
         }
@@ -1486,19 +1625,13 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
                 goto network_pkt_end;
             }
         }
-#if HAVE_NS_INITPARSE && HAVE_NS_PARSERR && HAVE_NS_SPRINTRR
-        if (!EMPTY(myregexes)) {
-            int     match, negmatch;
-            ns_msg  msg;
-            ns_sect s;
-
-            match    = -1;
-            negmatch = -1;
-            if (ns_initparse(dnspkt, dnslen, &msg) < 0) {
+        if (!EMPTY(myregexes) || match_qtype || nmatch_qtype) {
+            ldns_pkt* lpkt = 0;
+            if (ldns_wire2pkt(&lpkt, dnspkt, dnslen) != LDNS_STATUS_OK) {
                 /* DNS message may have padding, try get actual size */
-                if (errno == EMSGSIZE) {
-                    size_t dnslen2 = calcdnslen(dnspkt, dnslen);
-                    if (dnslen2 > 0 && dnslen2 < dnslen && ns_initparse(dnspkt, dnslen2, &msg) < 0) {
+                size_t dnslen2 = calcdnslen(dnspkt, dnslen);
+                if (dnslen2 > 0 && dnslen2 < dnslen) {
+                    if (ldns_wire2pkt(&lpkt, dnspkt, dnslen2) != LDNS_STATUS_OK) {
                         tcpstate_discard(tcpstate, "failed parse");
                         goto network_pkt_end;
                     }
@@ -1507,74 +1640,19 @@ void network_pkt(const char* descr, my_bpftimeval ts, unsigned pf,
                     goto network_pkt_end;
                 }
             }
-            /* Look at each section of the message:
-                 question, answer, authority, additional */
-            for (s = ns_s_qd; s < ns_s_max; s++) {
-                char        pres[SNAPLEN * 4];
-                const char* look;
-                int         count, n;
-                ns_rr       rr;
-
-                /* Look at each RR in the section (or each QNAME in
-                   the question section). */
-                count = ns_msg_count(msg, s);
-                for (n = 0; n < count; n++) {
-                    myregex_ptr myregex;
-
-                    if (ns_parserr(&msg, s, n, &rr) < 0) {
-                        tcpstate_discard(tcpstate, "failed parse");
-                        goto network_pkt_end;
-                    }
-                    if (s == ns_s_qd) {
-                        look = ns_rr_name(rr);
-                    } else {
-                        if (ns_sprintrr(&msg, &rr, NULL, ".",
-                                pres, sizeof pres)
-                            < 0) {
-                            tcpstate_discard(tcpstate, "failed parse");
-                            goto network_pkt_end;
-                        }
-                        look = pres;
-                    }
-                    for (myregex = HEAD(myregexes);
-                         myregex != NULL;
-                         myregex = NEXT(myregex, link)) {
-                        if (myregex->not ) {
-                            if (negmatch < 0)
-                                negmatch = 0;
-                        } else {
-                            if (match < 0)
-                                match = 0;
-                        }
-
-                        if (regexec(&myregex->reg, look, 0, NULL, 0) == 0) {
-                            if (myregex->not )
-                                negmatch++;
-                            else
-                                match++;
-
-                            if (dumptrace >= 2)
-                                fprintf(stderr,
-                                    "; \"%s\" %s~ /%s/ %d %d\n",
-                                    look,
-                                    myregex->not ? "!" : "",
-                                    myregex->str,
-                                    match,
-                                    negmatch);
-                        }
-                    }
-                }
-            }
-            /*
-             * Fail if any negative matching or if no match, match can be -1 which
-             * indicates that there are only negative matching
-             */
-            if (negmatch > 0 || match == 0) {
-                tcpstate_discard(tcpstate, "failed regex match");
+            char* reason = 0;
+            if ((match_qtype || nmatch_qtype) && _filter_by_qtype(lpkt, &reason)) {
+                ldns_pkt_free(lpkt);
+                tcpstate_discard(tcpstate, reason);
                 goto network_pkt_end;
             }
+            if (!EMPTY(myregexes) && _filter_by_qname(lpkt, &reason)) {
+                ldns_pkt_free(lpkt);
+                tcpstate_discard(tcpstate, reason);
+                goto network_pkt_end;
+            }
+            ldns_pkt_free(lpkt);
         }
-#endif /* HAVE_NS_INITPARSE && HAVE_NS_PARSERR && HAVE_NS_SPRINTRR */
 
         /* Policy hiding. */
         if (end_hide != 0) {
@@ -1658,10 +1736,10 @@ uint16_t in_checksum(const u_char* ptr, size_t len)
     unsigned sum = 0, top;
 
     /* Main body. */
-    while (len >= NS_INT16SZ) {
+    while (len >= 2) {
         sum += *(const uint16_t*)ptr;
-        ptr += NS_INT16SZ;
-        len -= NS_INT16SZ;
+        ptr += 2;
+        len -= 2;
     }
 
     /* Leftover octet? */
